@@ -1,8 +1,8 @@
-import os
-import io
 import functools
+import io
 import logging
 import math
+import os
 
 logger = logging.getLogger("fsspec")
 
@@ -58,8 +58,8 @@ class MMapCache(BaseCache):
         self.cache = self._makefile()
 
     def _makefile(self):
-        import tempfile
         import mmap
+        import tempfile
 
         if self.size == 0:
             return bytearray()
@@ -80,6 +80,7 @@ class MMapCache(BaseCache):
         return mmap.mmap(fd.fileno(), self.size)
 
     def _fetch(self, start, end):
+        logger.debug(f"MMap cache fetching {start}-{end}")
         if start is None:
             start = 0
         if end is None:
@@ -95,6 +96,7 @@ class MMapCache(BaseCache):
             i = need.pop(0)
             sstart = i * self.blocksize
             send = min(sstart + self.blocksize, self.size)
+            logger.debug(f"MMap get block #{i} ({sstart}-{send}")
             self.cache[sstart:send] = self.fetcher(sstart, send)
             self.blocks.add(i)
 
@@ -154,6 +156,37 @@ class ReadAheadCache(BaseCache):
         return part + self.cache[:l]
 
 
+class FirstChunkCache(BaseCache):
+    """Caches the first block of a file only
+
+    This may be useful for file types where the metadata is stored in the header,
+    but is randomly accessed.
+    """
+
+    name = "first"
+
+    def __init__(self, blocksize, fetcher, size):
+        super().__init__(blocksize, fetcher, size)
+        self.cache = None
+
+    def _fetch(self, start, end):
+        start = start or 0
+        end = end or self.size
+        if start < self.blocksize:
+            if self.cache is None:
+                if end > self.blocksize:
+                    data = self.fetcher(0, end)
+                    self.cache = data[: self.blocksize]
+                    return data[start:]
+                self.cache = self.fetcher(0, self.blocksize)
+            part = self.cache[start:end]
+            if end > self.blocksize:
+                part += self.fetcher(self.blocksize, end)
+            return part
+        else:
+            return self.fetcher(start, end)
+
+
 class BlockCache(BaseCache):
     """
     Cache holding memory as a set of blocks.
@@ -389,11 +422,39 @@ class AllBytes(BaseCache):
         return self.data[start:end]
 
 
+class KnownPartsOfAFile(BaseCache):
+    name = "parts"
+
+    def __init__(self, blocksize, fetcher, size, data={}, **_):
+        super(KnownPartsOfAFile, self).__init__(blocksize, fetcher, size)
+
+        # simple consolidation of contiguous blocks
+        for start0, stop in data.copy():
+            for start, stop1 in data.copy():
+                if stop == start:
+                    data[(start0, stop1)] = data.pop((start0, stop)) + data.pop(
+                        (start, stop1)
+                    )
+        self.data = data
+
+    def _fetch(self, start, stop):
+        for (loc0, loc1), data in self.data.items():
+            if loc0 <= start < loc1 and loc0 <= stop <= loc1:
+                off = start - loc0
+                out = data[off : off + stop - start]
+                # reads beyond buffer are padded with zero
+                out += b"\x00" * (stop - start - len(out))
+                return out
+        raise ValueError("Read outside of know parts of file")
+
+
 caches = {
     "none": BaseCache,
     "mmap": MMapCache,
     "bytes": BytesCache,
     "readahead": ReadAheadCache,
     "block": BlockCache,
+    "first": FirstChunkCache,
     "all": AllBytes,
+    "parts": KnownPartsOfAFile,
 }
