@@ -8,6 +8,7 @@ import time
 from shutil import move, rmtree
 
 from fsspec import AbstractFileSystem, filesystem
+from fsspec.callbacks import _DEFAULT_CALLBACK
 from fsspec.compression import compr
 from fsspec.core import BaseCache, MMapCache
 from fsspec.spec import AbstractBufferedFile
@@ -160,6 +161,7 @@ class CachingFileSystem(AbstractFileSystem):
                     else:
                         c["blocks"] = set(c["blocks"]).union(cache[k]["blocks"])
                     c["time"] = max(c["time"], cache[k]["time"])
+                    c["uid"] = cache[k]["uid"]
 
             # Files can be added to cache after it was written once
             for k, c in cache.items():
@@ -289,6 +291,7 @@ class CachingFileSystem(AbstractFileSystem):
             fn = os.path.join(self.storage[-1], hash)
             blocks = set()
             detail = {
+                "original": path,
                 "fn": hash,
                 "blocks": blocks,
                 "time": time.time(),
@@ -340,7 +343,7 @@ class CachingFileSystem(AbstractFileSystem):
         path = self._strip_protocol(f.path)
 
         c = self.cached_files[-1][path]
-        if c["blocks"] is not True and len(["blocks"]) * f.blocksize >= f.size:
+        if c["blocks"] is not True and len(c["blocks"]) * f.blocksize >= f.size:
             c["blocks"] = True
         try:
             logger.debug("going to save")
@@ -351,6 +354,7 @@ class CachingFileSystem(AbstractFileSystem):
         except NameError:
             logger.debug("Cache save failed due to interpreter shutdown")
         close()
+        f.closed = True
 
     def __getattribute__(self, item):
         if item in [
@@ -377,13 +381,19 @@ class CachingFileSystem(AbstractFileSystem):
             "_mkcache",
             "local_file",
             "_paths_from_path",
+            "get_mapper",
             "open_many",
             "commit_many",
             "hash_name",
+            "__hash__",
+            "__eq__",
+            "to_json",
         ]:
             # all the methods defined in this class. Note `open` here, since
             # it calls `_open`, but is actually in superclass
-            return lambda *args, **kw: getattr(type(self), item)(self, *args, **kw)
+            return lambda *args, **kw: getattr(type(self), item).__get__(self)(
+                *args, **kw
+            )
         if item in ["__reduce_ex__"]:
             raise AttributeError
         if item in ["_cache"]:
@@ -402,7 +412,7 @@ class CachingFileSystem(AbstractFileSystem):
             # attributed belonging to the target filesystem
             cls = type(fs)
             m = getattr(cls, item)
-            if inspect.isfunction(m) and (
+            if (inspect.isfunction(m) or inspect.isdatadescriptor(m)) and (
                 not hasattr(m, "__self__") or m.__self__ is None
             ):
                 # instance method
@@ -412,6 +422,45 @@ class CachingFileSystem(AbstractFileSystem):
             # attributes of the superclass, while target is being set up
             return super().__getattribute__(item)
 
+    def __eq__(self, other):
+        """Test for equality."""
+        if self is other:
+            return True
+        if not isinstance(other, type(self)):
+            return False
+        return (
+            self.storage == other.storage
+            and self.kwargs == other.kwargs
+            and self.cache_check == other.cache_check
+            and self.check_files == other.check_files
+            and self.expiry == other.expiry
+            and self.compression == other.compression
+            and self.same_names == other.same_names
+            and self.target_protocol == other.target_protocol
+        )
+
+    def __hash__(self):
+        """Calculate hash."""
+        return (
+            hash(tuple(self.storage))
+            ^ hash(str(self.kwargs))
+            ^ hash(self.cache_check)
+            ^ hash(self.check_files)
+            ^ hash(self.expiry)
+            ^ hash(self.compression)
+            ^ hash(self.same_names)
+            ^ hash(self.target_protocol)
+        )
+
+    def to_json(self):
+        """Calculate JSON representation.
+
+        Not implemented yet for CachingFileSystem.
+        """
+        raise NotImplementedError(
+            "CachingFileSystem JSON representation not implemented"
+        )
+
 
 class WholeFileCacheFileSystem(CachingFileSystem):
     """Caches whole remote files on first access
@@ -456,6 +505,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
             # update metadata - only happens when downloads are successful
             newdetail = [
                 {
+                    "original": path,
                     "fn": self.hash_name(path, self.same_names),
                     "blocks": True,
                     "time": time.time(),
@@ -484,6 +534,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
         hash = self.hash_name(path, self.same_names)
         fn = os.path.join(self.storage[-1], hash)
         detail = {
+            "original": path,
             "fn": hash,
             "blocks": True,
             "time": time.time(),
@@ -493,7 +544,14 @@ class WholeFileCacheFileSystem(CachingFileSystem):
         logger.debug("Copying %s to local cache" % path)
         return fn
 
-    def cat(self, path, recursive=False, on_error="raise", **kwargs):
+    def cat(
+        self,
+        path,
+        recursive=False,
+        on_error="raise",
+        callback=_DEFAULT_CALLBACK,
+        **kwargs,
+    ):
         paths = self.expand_path(
             path, recursive=recursive, maxdepth=kwargs.get("maxdepth", None)
         )
@@ -512,7 +570,12 @@ class WholeFileCacheFileSystem(CachingFileSystem):
         if getpaths:
             self.fs.get(getpaths, storepaths)
             self.save_cache()
-        out = {path: open(fn, "rb").read() for path, fn in zip(paths, fns)}
+
+        out = {}
+        callback.set_size(len(paths))
+        for p, fn in zip(paths, fns):
+            out[p] = open(fn, "rb").read()
+            callback.relative_update(1)
         if isinstance(path, str) and len(paths) == 1 and recursive is False:
             out = out[paths[0]]
         return out
@@ -527,7 +590,15 @@ class WholeFileCacheFileSystem(CachingFileSystem):
             _, blocks = detail["fn"], detail["blocks"]
             if blocks is True:
                 logger.debug("Opening local copy of %s" % path)
-                return open(fn, mode)
+
+                # In order to support downstream filesystems to be able to
+                # infer the compression from the original filename, like
+                # the `TarFileSystem`, let's extend the `io.BufferedReader`
+                # fileobject protocol by adding a dedicated attribute
+                # `original`.
+                f = open(fn, mode)
+                f.original = detail.get("original")
+                return f
             else:
                 raise ValueError(
                     "Attempt to open partially cached file %s"
