@@ -1,16 +1,20 @@
 import io
 import logging
 import os
+import threading
 import warnings
+import weakref
 from distutils.version import LooseVersion
 from errno import ESPIPE
 from glob import has_magic
 from hashlib import sha256
 
-from .config import apply_config
+from .callbacks import _DEFAULT_CALLBACK
+from .config import apply_config, conf
 from .dircache import DirCache
 from .transaction import Transaction
 from .utils import (
+    _unstrip_protocol,
     get_package_version_without_import,
     other_paths,
     read_block,
@@ -47,7 +51,11 @@ class _Cached(type):
         # Note: we intentionally create a reference here, to avoid garbage
         # collecting instances when all other references are gone. To really
         # delete a FileSystem, the cache must be cleared.
-        cls._cache = {}
+        if conf.get("weakref_instance_cache"):  # pragma: no cover
+            # debug option for analysing fork/spawn conditions
+            cls._cache = weakref.WeakValueDictionary()
+        else:
+            cls._cache = {}
         cls._pid = os.getpid()
 
     def __call__(cls, *args, **kwargs):
@@ -55,7 +63,9 @@ class _Cached(type):
         extra_tokens = tuple(
             getattr(cls, attr, None) for attr in cls._extra_tokenize_attributes
         )
-        token = tokenize(cls, *args, *extra_tokens, **kwargs)
+        token = tokenize(
+            cls, cls._pid, threading.get_ident(), *args, *extra_tokens, **kwargs
+        )
         skip = kwargs.pop("skip_instance_cache", False)
         if os.getpid() != cls._pid:
             cls._cache.clear()
@@ -249,7 +259,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         # Not necessary to implement invalidation mechanism, may have no cache.
         # But if have, you should call this method of parent class from your
         # subclass to ensure expiring caches after transacations correctly.
-        # See the implementaion of FTPFileSystem in ftp.py
+        # See the implementation of FTPFileSystem in ftp.py
         if self._intrans:
             self._invalidated_caches_in_transaction.append(path)
 
@@ -340,7 +350,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """
         parent = self._parent(path)
         if path.rstrip("/") in self.dircache:
-            return self.dircache[path]
+            return self.dircache[path.rstrip("/")]
         try:
             files = [
                 f
@@ -382,7 +392,9 @@ class AbstractFileSystem(up, metaclass=_Cached):
         try:
             listing = self.ls(path, detail=True, **kwargs)
         except (FileNotFoundError, IOError):
-            return [], [], []
+            if detail:
+                return path, {}, {}
+            return path, [], []
 
         for info in listing:
             # each info name must be at least [path]/part , but here
@@ -435,7 +447,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
             if withdirs:
                 files.update(dirs)
             out.update({info["name"]: info for name, info in files.items()})
-        if self.isfile(path) and path not in out:
+        if not out and self.isfile(path):
             # walk works on directories, but find should also return [path]
             # when path happens to be a file
             out[path] = {}
@@ -568,6 +580,11 @@ class AbstractFileSystem(up, metaclass=_Cached):
             # any exception allowed bar FileNotFoundError?
             return False
 
+    def lexists(self, path, **kwargs):
+        """If there is a file at the given path (including
+        broken links)"""
+        return self.exists(path)
+
     def info(self, path, **kwargs):
         """Give details of entry at path
 
@@ -634,12 +651,27 @@ class AbstractFileSystem(up, metaclass=_Cached):
             return False
 
     def cat_file(self, path, start=None, end=None, **kwargs):
-        """ Get the content of a file """
+        """Get the content of a file
+
+        Parameters
+        ----------
+        path: URL of file on this filesystems
+        start, end: int
+            Bytes limits of the read. If negative, backwards from end,
+            like usual python slices. Either can be None for start or
+            end of file, respectively
+        kwargs: passed to ``open()``.
+        """
         # explicitly set buffering off?
         with self.open(path, "rb", **kwargs) as f:
             if start is not None:
-                f.seek(start)
+                if start >= 0:
+                    f.seek(start)
+                else:
+                    f.seek(max(0, f.size + start))
             if end is not None:
+                if end < 0:
+                    end = f.size + end
                 return f.read(end - f.tell())
             return f.read()
 
@@ -652,6 +684,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Put value into path
 
         (counterpart to ``cat``)
+
         Parameters
         ----------
         path: string or dict(str, bytes)
@@ -669,18 +702,39 @@ class AbstractFileSystem(up, metaclass=_Cached):
         else:
             raise ValueError("path must be str or dict")
 
+    def cat_ranges(self, paths, starts, ends, max_gap=None, **kwargs):
+        if max_gap is not None:
+            raise NotImplementedError
+        if not isinstance(paths, list):
+            raise TypeError
+        if not isinstance(starts, list):
+            starts = [starts] * len(paths)
+        if not isinstance(ends, list):
+            ends = [starts] * len(paths)
+        if len(starts) != len(paths) or len(ends) != len(paths):
+            raise ValueError
+        return [self.cat_file(p, s, e) for p, s, e in zip(paths, starts, ends)]
+
     def cat(self, path, recursive=False, on_error="raise", **kwargs):
         """Fetch (potentially multiple) paths' contents
 
-        Returns a dict of {path: contents} if there are multiple paths
-        or the path has been otherwise expanded
-
+        Parameters
+        ----------
+        recursive: bool
+            If True, assume the path(s) are directories, and get all the
+            contained files
         on_error : "raise", "omit", "return"
             If raise, an underlying exception will be raised (converted to KeyError
             if the type is in self.missing_exceptions); if omit, keys with exception
             will simply not be included in the output; if "return", all keys are
             included in the output, but the value will be bytes or an exception
             instance.
+        kwargs: passed to cat_file
+
+        Returns
+        -------
+        dict of {path: contents} if there are multiple paths
+        or the path has been otherwise expanded
         """
         paths = self.expand_path(path, recursive=recursive)
         if (
@@ -699,21 +753,24 @@ class AbstractFileSystem(up, metaclass=_Cached):
                         out[path] = e
             return out
         else:
-            return self.cat_file(paths[0])
+            return self.cat_file(paths[0], **kwargs)
 
-    def get_file(self, rpath, lpath, **kwargs):
+    def get_file(self, rpath, lpath, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy single remote file to local"""
         if self.isdir(rpath):
             os.makedirs(lpath, exist_ok=True)
-        else:
-            with self.open(rpath, "rb", **kwargs) as f1:
-                with open(lpath, "wb") as f2:
-                    data = True
-                    while data:
-                        data = f1.read(self.blocksize)
-                        f2.write(data)
+            return None
 
-    def get(self, rpath, lpath, recursive=False, **kwargs):
+        with self.open(rpath, "rb", **kwargs) as f1:
+            callback.set_size(getattr(f1, "size", None))
+            with open(lpath, "wb") as f2:
+                data = True
+                while data:
+                    data = f1.read(self.blocksize)
+                    segment_len = f2.write(data)
+                    callback.relative_update(segment_len)
+
+    def get(self, rpath, lpath, recursive=False, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy file(s) to local.
 
         Copies a specific file or tree of files (if recursive=True). If lpath
@@ -729,23 +786,31 @@ class AbstractFileSystem(up, metaclass=_Cached):
             lpath = make_path_posix(lpath)
         rpaths = self.expand_path(rpath, recursive=recursive)
         lpaths = other_paths(rpaths, lpath)
-        for lpath, rpath in zip(lpaths, rpaths):
+
+        callback.set_size(len(lpaths))
+        for lpath, rpath in callback.wrap(zip(lpaths, rpaths)):
+            callback.branch(rpath, lpath, kwargs)
             self.get_file(rpath, lpath, **kwargs)
 
-    def put_file(self, lpath, rpath, **kwargs):
+    def put_file(self, lpath, rpath, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy single file to remote"""
         if os.path.isdir(lpath):
             self.makedirs(rpath, exist_ok=True)
-        else:
-            with open(lpath, "rb") as f1:
-                self.mkdirs(os.path.dirname(rpath), exist_ok=True)
-                with self.open(rpath, "wb", **kwargs) as f2:
-                    data = True
-                    while data:
-                        data = f1.read(self.blocksize)
-                        f2.write(data)
+            return None
 
-    def put(self, lpath, rpath, recursive=False, **kwargs):
+        with open(lpath, "rb") as f1:
+            callback.set_size(f1.seek(0, 2))
+            f1.seek(0)
+
+            self.mkdirs(self._parent(os.fspath(rpath)), exist_ok=True)
+            with self.open(rpath, "wb", **kwargs) as f2:
+                data = True
+                while data:
+                    data = f1.read(self.blocksize)
+                    segment_len = f2.write(data)
+                    callback.relative_update(segment_len)
+
+    def put(self, lpath, rpath, recursive=False, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy file(s) from local.
 
         Copies a specific file or tree of files (if recursive=True). If rpath
@@ -765,9 +830,13 @@ class AbstractFileSystem(up, metaclass=_Cached):
             lpath = make_path_posix(lpath)
         fs = LocalFileSystem()
         lpaths = fs.expand_path(lpath, recursive=recursive)
-        rpaths = other_paths(lpaths, rpath)
+        rpaths = other_paths(
+            lpaths, rpath, exists=isinstance(rpath, str) and self.isdir(rpath)
+        )
 
-        for lpath, rpath in zip(lpaths, rpaths):
+        callback.set_size(len(rpaths))
+        for lpath, rpath in callback.wrap(zip(lpaths, rpaths)):
+            callback.branch(lpath, rpath, kwargs)
             self.put_file(lpath, rpath, **kwargs)
 
     def head(self, path, size=1024):
@@ -1076,7 +1145,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
     def get_mapper(self, root, check=False, create=False):
         """Create key/value store based on this file-system
 
-        Makes a MutibleMapping interface to the FS at the given root path.
+        Makes a MutableMapping interface to the FS at the given root path.
         See ``fsspec.mapping.FSMap`` for further details.
         """
         from .mapping import FSMap
@@ -1110,47 +1179,47 @@ class AbstractFileSystem(up, metaclass=_Cached):
     # Aliases
 
     def makedir(self, path, create_parents=True, **kwargs):
-        """Alias of :ref:`FilesystemSpec.mkdir`."""
+        """Alias of `AbstractFileSystem.mkdir`."""
         return self.mkdir(path, create_parents=create_parents, **kwargs)
 
     def mkdirs(self, path, exist_ok=False):
-        """Alias of :ref:`FilesystemSpec.makedirs`."""
+        """Alias of `AbstractFileSystem.makedirs`."""
         return self.makedirs(path, exist_ok=exist_ok)
 
     def listdir(self, path, detail=True, **kwargs):
-        """Alias of :ref:`FilesystemSpec.ls`."""
+        """Alias of `AbstractFileSystem.ls`."""
         return self.ls(path, detail=detail, **kwargs)
 
     def cp(self, path1, path2, **kwargs):
-        """Alias of :ref:`FilesystemSpec.copy`."""
+        """Alias of `AbstractFileSystem.copy`."""
         return self.copy(path1, path2, **kwargs)
 
     def move(self, path1, path2, **kwargs):
-        """Alias of :ref:`FilesystemSpec.mv`."""
+        """Alias of `AbstractFileSystem.mv`."""
         return self.mv(path1, path2, **kwargs)
 
     def stat(self, path, **kwargs):
-        """Alias of :ref:`FilesystemSpec.info`."""
+        """Alias of `AbstractFileSystem.info`."""
         return self.info(path, **kwargs)
 
     def disk_usage(self, path, total=True, maxdepth=None, **kwargs):
-        """Alias of :ref:`FilesystemSpec.du`."""
+        """Alias of `AbstractFileSystem.du`."""
         return self.du(path, total=total, maxdepth=maxdepth, **kwargs)
 
     def rename(self, path1, path2, **kwargs):
-        """Alias of :ref:`FilesystemSpec.mv`."""
+        """Alias of `AbstractFileSystem.mv`."""
         return self.mv(path1, path2, **kwargs)
 
     def delete(self, path, recursive=False, maxdepth=None):
-        """Alias of :ref:`FilesystemSpec.rm`."""
+        """Alias of `AbstractFileSystem.rm`."""
         return self.rm(path, recursive=recursive, maxdepth=maxdepth)
 
     def upload(self, lpath, rpath, recursive=False, **kwargs):
-        """Alias of :ref:`FilesystemSpec.put`."""
+        """Alias of `AbstractFileSystem.put`."""
         return self.put(lpath, rpath, recursive=recursive, **kwargs)
 
     def download(self, rpath, lpath, recursive=False, **kwargs):
-        """Alias of :ref:`FilesystemSpec.get`."""
+        """Alias of `AbstractFileSystem.get`."""
         return self.get(rpath, lpath, recursive=recursive, **kwargs)
 
     def sign(self, path, expiration=100, **kwargs):
@@ -1173,14 +1242,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
         Raises
         ------
-        NotImplementedError : if method is not implemented for a fileystem
+        NotImplementedError : if method is not implemented for a filesystem
         """
         raise NotImplementedError("Sign is not implemented for this filesystem")
 
     def _isfilestore(self):
         # Originally inherited from pyarrow DaskFileSystem. Keeping this
         # here for backwards compatibility as long as pyarrow uses its
-        # legacy ffspec-compatible filesystems and thus accepts fsspec
+        # legacy fsspec-compatible filesystems and thus accepts fsspec
         # filesystems as well
         return False
 
@@ -1191,10 +1260,11 @@ class AbstractBufferedFile(io.IOBase):
     In the case that the backend does not provide a pythonic file-like object
     already, this class contains much of the logic to build one. The only
     methods that need to be overridden are ``_upload_chunk``,
-    ``_initate_upload`` and ``_fetch_range``.
+    ``_initiate_upload`` and ``_fetch_range``.
     """
 
     DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
+    _details = None
 
     def __init__(
         self,
@@ -1205,6 +1275,7 @@ class AbstractBufferedFile(io.IOBase):
         autocommit=True,
         cache_type="readahead",
         cache_options=None,
+        size=None,
         **kwargs,
     ):
         """
@@ -1228,6 +1299,8 @@ class AbstractBufferedFile(io.IOBase):
         cache_options : dict
             Additional options passed to the constructor for the cache specified
             by `cache_type`.
+        size: int
+            If given and in read mode, suppressed having to look up the file size
         kwargs:
             Gets stored as self.kwargs
         """
@@ -1261,9 +1334,10 @@ class AbstractBufferedFile(io.IOBase):
         if mode not in {"ab", "rb", "wb"}:
             raise NotImplementedError("File mode not supported")
         if mode == "rb":
-            if not hasattr(self, "details"):
-                self.details = fs.info(path)
-            self.size = self.details["size"]
+            if size is not None:
+                self.size = size
+            else:
+                self.size = self.details["size"]
             self.cache = caches[cache_type](
                 self.blocksize, self._fetch_range, self.size, **cache_options
             )
@@ -1273,6 +1347,21 @@ class AbstractBufferedFile(io.IOBase):
             self.forced = False
             self.location = None
 
+    @property
+    def details(self):
+        if self._details is None:
+            self._details = self.fs.info(self.path)
+        return self._details
+
+    @details.setter
+    def details(self, value):
+        self._details = value
+        self.size = value["size"]
+
+    @property
+    def full_name(self):
+        return _unstrip_protocol(self.path, self.fs)
+
     @property
     def closed(self):
         # get around this attr being read-only in IOBase
@@ -1392,7 +1481,11 @@ class AbstractBufferedFile(io.IOBase):
         if self.offset is None:
             # Initialize a multipart upload
             self.offset = 0
-            self._initiate_upload()
+            try:
+                self._initiate_upload()
+            except:  # noqa: E722
+                self.closed = True
+                raise
 
         if self._upload_chunk(final=force) is not False:
             self.offset += self.buffer.seek(0, 2)
@@ -1515,6 +1608,8 @@ class AbstractBufferedFile(io.IOBase):
 
         Finalizes writes, discards cache
         """
+        if getattr(self, "_unclosable", False):
+            return
         if self.closed:
             return
         if self.mode == "rb":
